from ..chunks import *
from ..utils.reader import Reader


def _(*args):
    print('[ScwUtils]', end=' ')
    for arg in args:
        print(arg, end=' ')
    print()


class Parser(Reader):
    def __init__(self, file_data: bytes):
        super().__init__(file_data)
        self.file_data = file_data
        self.parsed = {
            'header': {},
            'materials': [],
            'geometries': [],
            'cameras': [],
            'nodes': []
        }
        self.chunks = []

        file_magic = self.read(4)
        if file_magic != b'SC3D':
            raise TypeError('File Magic isn\'t "SC3D"')

    def split_chunks(self):
        # len(Chunk Length) + len(Chunk Name) + len(Chunk CRC)
        while len(self.file_data[self.tell():]) >= 12:
            chunk_length = self.readUInt32()
            chunk_name = self.readChar(4)
            chunk_data = self.read(chunk_length)
            chunk_crc = self.readUInt32()

            self.chunks.append({
                'name': chunk_name,
                'data': chunk_data,
                'crc': chunk_crc
            })

    def parse(self):
        for chunk in self.chunks:
            chunk_name = chunk['name']
            chunk_data = chunk['data']

            if chunk_name == 'HEAD':
                head = HEAD.Decoder(chunk_data)
                self.parsed['header'] = head.readed
            elif chunk_name == 'MATE':
                mate = MATE.Decoder(chunk_data, self.parsed['header'])
                self.parsed['materials'].append(mate.readed)
            elif chunk_name == 'GEOM':
                geom = GEOM.Decoder(chunk_data, self.parsed['header'])
                self.parsed['geometries'].append(geom.readed)
            elif chunk_name == 'CAME':
                came = CAME.Decoder(chunk_data, self.parsed['header'])
                self.parsed['cameras'].append(came.readed)
            elif chunk_name == 'NODE':
                node = NODE.Decoder(chunk_data, self.parsed['header'])
                self.parsed['nodes'] = node.readed
            elif chunk_name == 'WEND':
                WEND.Decoder(chunk_data)
            else:
                raise TypeError(f'Unknown chunk: {chunk_name}')



from io import BufferedReader, BytesIO
from struct import unpack


class Reader(BufferedReader):
    def __init__(self, initial_bytes: bytes):
        super(Reader, self).__init__(BytesIO(initial_bytes))

    def readUInt64(self):
        return unpack('>Q', self.read(8))[0]

    def readInt64(self):
        return unpack('>q', self.read(8))[0]

    def readFloat(self):
        return unpack('>f', self.read(4))[0]

    def readUInt32(self):
        return unpack('>I', self.read(4))[0]

    def readInt32(self):
        return unpack('>i', self.read(4))[0]

    def readUInt16(self):
        return unpack('>H', self.read(2))[0]

    def readInt16(self):
        return unpack('>h', self.read(2))[0]

    def readUInt8(self):
        return unpack('>B', self.read(1))[0]

    def readInt8(self):
        return unpack('>b', self.read(1))[0]

    def readBool(self):
        return unpack('>?', self.read(1))[0]

    def readUInteger(self, length: int = 1):
        return int.from_bytes(self.read(length), 'big', signed=False)

    def readInteger(self, length: int = 1):
        return int.from_bytes(self.read(length), 'big', signed=True)

    readULong = readUInt64
    readLong = readInt64

    readUShort = readUInt16
    readShort = readInt16

    readUByte = readUInt8
    readByte = readInt8

    readUInt = readUInteger
    readInt = readInteger

    def readChar(self, length: int = 1):
        return self.read(length).decode('utf-8')

    def readString(self):
        return self.readChar(self.readUShort())



from ..utils.reader import Reader
from ..utils.writer import Writer


class Decoder(Reader):
    def __init__(self, initial_bytes: bytes):
        super().__init__(initial_bytes)

        # Variables
        self.readed = {}
        # Variables

        version = self.readUShort()
        frame_rate = self.readUShort()
        self.readUShort()
        self.readUShort()
        materials_file = self.readString()
        self.readUByte()

        self.readed['version'] = version
        self.readed['frame_rate'] = frame_rate
        self.readed['materials_file'] = materials_file


class Encoder(Writer):
    def __init__(self, data: dict):
        super().__init__()
        self.name = 'HEAD'
        self.data = data

        self.encode()

        self.length = len(self.buffer)

    def encode(self):
        self.writeUShort(self.data['version'])  # version
        self.writeUShort(self.data['frame_rate'])  # frame rate
        self.writeUShort(0)
        self.writeUShort(249)
        self.writeString(self.data['materials_file'])  # materials file
        self.writeUByte(0)



from ..utils.reader import Reader
from ..utils.writer import Writer


class Decoder(Reader):
    def __init__(self, initial_bytes: bytes, header: dict):
        super().__init__(initial_bytes)

        # Variables
        self.readed = []
        # Variables

        nodes_count = self.readUShort()
        for node in range(nodes_count):
            self.readed.append({})
            name = self.readString()
            parent = self.readString()
            self.readed[node] = {'name': name,
                                 'parent': parent}

            has_target = True if self.readUShort() else False
            self.readed[node]['has_target'] = has_target
            if has_target:
                target_type = self.readChar(4)
                target_name = self.readString()
                binds_count = self.readUShort()
                binds = []
                for bind in range(binds_count):
                    binds.append({})
                    symbol = self.readString()
                    target = self.readString()
                    binds[bind] = {'symbol': symbol,
                                   'target': target}
                self.readed[node]['target_type'] = target_type
                self.readed[node]['target'] = target_name
                self.readed[node]['binds'] = binds

            frames_count = self.readUShort()
            self.readed[node]['frames'] = []
            if frames_count > 0:
                settings = list(bin(self.readUByte())[2:].zfill(8))
                self.readed[node]['frames_settings'] = settings
                for frame in range(frames_count):
                    self.readed[node]['frames'].append({})
                    frame_id = self.readUShort()
                    if settings[7] or frame == 0:  # Rotation
                        rot_x = self.readUShort() / 32512
                        rot_y = self.readUShort() / 32512
                        rot_z = self.readUShort() / 32512
                        w = self.readUShort() / 32512

                    if settings[4] or frame == 0:  # Position X
                        pos_x = self.readFloat()
                    if settings[5] or frame == 0:  # Position Y
                        pos_y = self.readFloat()
                    if settings[6] or frame == 0:  # Position Z
                        pos_z = self.readFloat()

                    if settings[1] or frame == 0:  # Scale X
                        scale_x = self.readFloat()
                    if settings[2] or frame == 0:  # Scale Y
                        scale_y = self.readFloat()
                    if settings[3] or frame == 0:  # Scale Z
                        scale_z = self.readFloat()
                    self.readed[node]['frames'][frame] = {'frame_id': frame_id,
                                                          'rotation': {'x': rot_x,
                                                                       'y': rot_y,
                                                                       'z': rot_z,
                                                                       'w': w},
                                                          'position': {'x': pos_x,
                                                                       'y': pos_y,
                                                                       'z': pos_z},
                                                          'scale': {'x': scale_x,
                                                                    'y': scale_y,
                                                                    'z': scale_z}}


class Encoder(Writer):
    def __init__(self, data: dict):
        super().__init__()
        self.name = 'NODE'
        self.data = data

        self.encode()

        self.length = len(self.buffer)

    def encode(self):
        self.writeUShort(len(self.data))
        for node in self.data:
            self.writeString(node['name'])
            self.writeString(node['parent'])

            self.writeUShort(1 if node['has_target'] else 0)
            if node['has_target']:
                self.writeChar(node['target_type'])
                self.writeString(node['target'])
                self.writeUShort(len(node['binds']))
                for bind in node['binds']:
                    self.writeString(bind['symbol'])
                    self.writeString(bind['target'])

            if 'frames_settings' in node:
                frames_settings = node['frames_settings']
            else:
                frames_settings = None
            self.encode_frames(node['frames'], frames_settings)

    def encode_frames(self, frames, frames_settings):
        self.writeUShort(len(frames))
        if len(frames) > 0:
            self.writeUByte(int(''.join([str(item) for item in frames_settings])[::], 2))
            for frame in frames:
                self.writeUShort(frame['frame_id'])
                if frames_settings[7] or frames.index(frame) == 0:  # Rotation
                    self.writeUShort(round(frame['rotation']['x'] * 32512))
                    self.writeUShort(round(frame['rotation']['y'] * 32512))
                    self.writeUShort(round(frame['rotation']['z'] * 32512))
                    self.writeUShort(round(frame['rotation']['w'] * 32512))

                if frames_settings[4] or frames.index(frame) == 0:  # Position X
                    self.writeFloat(frame['position']['x'])
                if frames_settings[5] or frames.index(frame) == 0:  # Position Y
                    self.writeFloat(frame['position']['y'])
                if frames_settings[6] or frames.index(frame) == 0:  # Position Z
                    self.writeFloat(frame['position']['z'])

                if frames_settings[1] or frames.index(frame) == 0:  # Scale X
                    self.writeFloat(frame['scale']['x'])
                if frames_settings[2] or frames.index(frame) == 0:  # Scale Y
                    self.writeFloat(frame['scale']['y'])
                if frames_settings[3] or frames.index(frame) == 0:  # Scale Z
                    self.writeFloat(frame['scale']['z'])



def _(*args):
    print('[ScwUtils]', end=' ')
    for arg in args:
        print(arg, end=' ')
    print()


class Writer:
    def __init__(self, data: dict):
        self.writen = ''

        temp_vertices_offsets = {
            'POSITION': 0,
            'TEXCOORD': 0,
            'NORMAL': 0
        }

        vertices_offsets = {
            'POSITION': 0,
            'TEXCOORD': 0,
            'NORMAL': 0
        }

        for geom in data['geometries']:
            for key in vertices_offsets.keys():
                vertices_offsets[key] = temp_vertices_offsets[key]
            prefix = ''

            name = geom['name']
            vertices = geom['vertices']
            materials = geom['materials']
            for vertex in vertices:
                if vertex['type'] == 'POSITION':
                    prefix = 'v '
                elif vertex['type'] == 'NORMAL':
                    prefix = 'vn '
                elif vertex['type'] == 'TEXCOORD':
                    prefix = 'vt '

                temp_vertices_offsets[vertex['type']] += len(vertex['vertex'])

                for item in vertex['vertex']:
                    temp_string = prefix
                    for subitem in item:
                        temp_string += str(subitem * vertex['scale']) + ' '
                    self.writen += f'{temp_string}\n'
                self.writen += '\n\n'
            for material in materials:
                self.writen += f'o {name}_{material["name"]}\n\n'
                for item in material['polygons']:
                    temp_string = 'f '
                    for subitem in item:
                        temp_list = [
                            str(subitem[0] + vertices_offsets['POSITION'] + 1),  # POSITION
                            str(subitem[2] + vertices_offsets['TEXCOORD'] + 1),  # TEXCOORD
                            str(subitem[1] + vertices_offsets['NORMAL'] + 1)  # NORMAL
                        ]

                        temp_string += '/'.join(temp_list) + ' '
                    self.writen += f'{temp_string}\n'
                self.writen += '\n\n'



from struct import pack


class Writer:
    def __init__(self):
        super(Writer, self).__init__()
        self.buffer = b''

    def writeUInt64(self, integer: int):
        self.buffer += pack('>Q', integer)

    def writeInt64(self, integer: int):
        self.buffer += pack('>q', integer)

    def writeFloat(self, floating: float):
        self.buffer += pack('>f', floating)

    def writeUInt32(self, integer: int):
        self.buffer += pack('>I', integer)

    def writeInt32(self, integer: int):
        self.buffer += pack('>i', integer)

    def writeUInt16(self, integer: int):
        self.buffer += pack('>H', integer)

    def writeInt16(self, integer: int):
        self.buffer += pack('>h', integer)

    def writeUInt8(self, integer: int):
        self.buffer += pack('>B', integer)

    def writeInt8(self, integer: int):
        self.buffer += pack('>b', integer)

    def writeBool(self, boolean: bool):
        self.buffer += pack('>?', boolean)

    def writeUInteger(self, integer: int, length: int = 1):
        return integer.to_bytes(length, 'big', signed=False)

    def writeInteger(self, integer: int, length: int = 1):
        return integer.to_bytes(length, 'big', signed=True)

    writeULong = writeUInt64
    writeLong = writeInt64

    writeUShort = writeUInt16
    writeShort = writeInt16

    writeUByte = writeUInt8
    writeByte = writeInt8

    writeUInt = writeUInteger
    writeInt = writeInteger

    def writeChar(self, string: str):
        for char in list(string):
            self.buffer += char.encode('utf-8')

    def writeString(self, string: str):
        encoded = string.encode('utf-8')
        self.writeUShort(len(encoded))
        self.buffer += encoded



from xml.etree.ElementTree import *

from ..utils.matrix.matrix4x4 import Matrix4x4


def _(*args):
    print('[ScwUtils]', end=' ')
    for arg in args:
        print(arg, end=' ')
    print()


class Parser:
    def node(self, nodes):
        nodes_list = []
        for node in nodes:
            instance_geometry = node.findall('collada:instance_geometry', self.namespaces)
            instance_controller = node.findall('collada:instance_controller', self.namespaces)

            children = self.node(node.findall('collada:node', self.namespaces))
            node_data = {
                'name': node.attrib['id'],
                'has_target': True if instance_geometry or instance_controller else False
            }

            if node_data['has_target']:
                binds = []

                if instance_geometry:
                    instance_geometry = instance_geometry[0]

                    bind_material = instance_geometry.find('collada:bind_material', self.namespaces)
                    technique_common = bind_material[0]

                    for instance_material in technique_common:
                        binds.append({
                            'symbol': instance_material.attrib['symbol'],
                            'target': instance_material.attrib['target'][1:]
                        })
                elif instance_controller:
                    instance_controller = instance_controller[0]

                    bind_material = instance_controller.find('collada:bind_material', self.namespaces)
                    technique_common = bind_material[0]

                    for instance_material in technique_common:
                        binds.append({
                            'symbol': instance_material.attrib['symbol'],
                            'target': instance_material.attrib['target'][1:]
                        })

                if instance_geometry:
                    node_data['target_type'] = 'GEOM'
                elif instance_controller:
                    node_data['target_type'] = 'CONT'

                if instance_geometry:
                    geometry_url = instance_geometry.attrib['url']
                    node_data['target'] = geometry_url[1:]
                elif instance_controller:
                    controller_url = instance_controller.attrib['url']
                    node_data['target'] = controller_url[1:]

                node_data['binds'] = binds

            matrix = node.findall('collada:matrix', self.namespaces)
            if matrix:
                matrix_data = matrix[0].text.split()
                matrix_data = [[float(value) for value in matrix_data[x:x + 4]] for x in range(0, len(matrix_data), 4)]

                node_data['matrix'] = matrix_data

            node_data['children'] = children

            nodes_list.append(node_data)

        return nodes_list

    def fix_nodes_list(self, nodes, parent: str = ''):
        for node in nodes:
            node_data = {
                'name': node['name'],
                'parent': parent,
                'has_target': node['has_target']
            }

            if node_data['has_target']:
                node_data['target_type'] = node['target_type']
                node_data['target'] = node['target']
                node_data['binds'] = node['binds']

            if not node_data['has_target']:
                node_data['frames_settings'] = [0, 0, 0, 0, 0, 0, 0, 0]
                node_data['frames'] = []

                if 'matrix' in node:
                    matrix = Matrix4x4(node['matrix'])

                    scale = matrix.get_scale()
                    position = matrix.get_position()

                    frame_data = {
                        'frame_id': 0,
                        'rotation': {'x': 0, 'y': 0, 'z': 0, 'w': 0},
                        'position': {'x': position[0], 'y': position[1], 'z': position[2]},
                        'scale': {'x': scale[0], 'y': scale[1], 'z': scale[2]}
                    }

                    node_data['frames'].append(frame_data)
            else:
                node_data['frames'] = []

            # node_data['frames'] = node['frames']
            self.parsed['nodes'].append(node_data)
            self.fix_nodes_list(node['children'], node['name'])

    def __init__(self, file_data):
        self.parsed = {'header': {'version': 2,
                                  'frame_rate': 30,
                                  'materials_file': 'sc3d/character_materials.scw'},
                       'materials': [],
                       'geometries': [],
                       'cameras': [],
                       'nodes': []}

        self.geometry_info = {}

        root = fromstring(file_data)
        # tree = parse(file_path)
        # root = tree.getroot()

        self.namespaces = {
            'collada': 'http://www.collada.org/2005/11/COLLADASchema'
        }

        # Libraries
        self.library_materials = root.find('./collada:library_materials', self.namespaces)
        self.library_effects = root.find('./collada:library_effects', self.namespaces)

        self.library_geometries = root.find('./collada:library_geometries', self.namespaces)
        self.library_controllers = root.find('./collada:library_controllers', self.namespaces)

        library_scenes = root.find('./collada:library_visual_scenes', self.namespaces)
        # Libraries

        for material in self.library_materials:
            material_name = material.attrib['id']

            instance_effect = material.find('collada:instance_effect', self.namespaces)
            if instance_effect is not None:
                effect_url = instance_effect.attrib['url'][1:]
                effect = self.library_effects.find(f'collada:effect[@id="{effect_url}"]', self.namespaces)

                if effect is not None:
                    profile = None
                    for item in effect:
                        if 'profile' in item.tag:
                            profile = item
                    technique = profile.find('collada:technique', self.namespaces)

                    emission_data = None
                    ambient_data = None
                    diffuse_data = None

                    emission = technique[0].find('collada:emission', self.namespaces)
                    ambient = technique[0].find('collada:ambient', self.namespaces)
                    diffuse = technique[0].find('collada:diffuse', self.namespaces)

                    if 'color' in emission[0].tag:
                        emission_data = [float(item) for item in emission[0].text.split()]
                        emission_data[3] *= 255
                    elif 'texture' in emission[0].tag:
                        # emission_data = emission[0].attrib['texture']
                        emission_data = '.'

                    if 'color' in ambient[0].tag:
                        ambient_data = [float(item) for item in ambient[0].text.split()]
                        ambient_data[3] *= 255
                    elif 'texture' in ambient[0].tag:
                        # ambient_data = ambient[0].attrib['texture']
                        ambient_data = '.'

                    if 'color' in diffuse[0].tag:
                        diffuse_data = [float(item) for item in diffuse[0].text.split()]
                        diffuse_data[3] *= 255
                    elif 'texture' in diffuse[0].tag:
                        # diffuse_data = diffuse[0].attrib['texture']
                        diffuse_data = '.'

                    material_data = {
                        'name': material_name,
                        'effect': {
                            'ambient': ambient_data,
                            'diffuse': diffuse_data,
                            'specular': '.',
                            'colorize': [255, 255, 255, 255],
                            'emission': emission_data,
                            'lightmaps': {
                                'diffuse': 'sc3d/diffuse_lightmap.png',
                                'specular': 'sc3d/specular_lightmap.png'
                            },
                            'tint': [0, 0, 0, 0]
                        }
                    }

                    self.parsed['materials'].append(material_data)

        instance_scene = root.find('./collada:scene', self.namespaces).find('collada:instance_visual_scene',
                                                                            self.namespaces)
        scene_url = instance_scene.attrib['url'][1:]
        scene = library_scenes.find(f'collada:visual_scene[@id="{scene_url}"]', self.namespaces)

        nodes = self.node(scene.findall('collada:node', self.namespaces))
        self.fix_nodes_list(nodes)

    def parse_nodes(self):
        nodes = self.parsed['nodes']
        for node in nodes:
            node: dict = node  # this line for fix "Expected type"
            if node['has_target']:
                controller = None
                geometry = None

                if node['target_type'] == 'CONT':
                    controller = self.library_controllers \
                        .find(f'collada:controller[@id="{node["target"]}"]', self.namespaces)

                    geometry_url = controller[0].attrib['source'][1:]
                    geometry = self.library_geometries \
                        .find(f'collada:geometry[@id="{geometry_url}"]', self.namespaces)
                elif node['target_type'] == 'GEOM':
                    geometry = self.library_geometries \
                        .find(f'collada:geometry[@id="{node["target"]}"]', self.namespaces)

                if geometry is not None:
                    self.geometry_info = {'name': '',
                                          'group': node['parent'],
                                          'vertices': [],
                                          'have_bind_matrix': False,
                                          'materials': []}
                    if controller is not None:
                        self.parse_controller(controller)

                    self.parse_geometry(geometry)

    def parse_controller(self, controller):
        self.geometry_info['have_bind_matrix'] = True

        skin = controller[0]

        bind_shape_matrix = skin.find('collada:bind_shape_matrix', self.namespaces).text
        bind_shape_matrix = [float(x) for x in bind_shape_matrix.split()]

        self.geometry_info['bind_matrix'] = bind_shape_matrix

        self.geometry_info['joints'] = []
        joints = skin.find('collada:joints', self.namespaces)
        joint_inputs = joints.findall('collada:input', self.namespaces)
        for _input in joint_inputs:
            # semantic = _input.attrib['semantic']
            source_url = _input.attrib['source']
            source = skin.find(f'collada:source[@id="{source_url[1:]}"]', self.namespaces)

            accessor = source.find('collada:technique_common/collada:accessor', self.namespaces)
            accessor_stride = int(accessor.attrib['stride'])
            accessor_source_url = accessor.attrib['source']
            accessor_source = source.find(f'collada:*[@id="{accessor_source_url[1:]}"]', self.namespaces)
            params = accessor.findall('collada:param', self.namespaces)

            for param in params:
                param_name = param.attrib['name']
                # param_type = param.attrib['type']

                if param_name == 'JOINT':
                    for name in accessor_source.text.split():
                        self.geometry_info['joints'].append({
                            'name': name
                        })

                if param_name == 'TRANSFORM':
                    for x in range(int(accessor_source.attrib['count']) // int(accessor_stride)):
                        matrix = []
                        for y in accessor_source.text.split()[x * accessor_stride:(x + 1) * accessor_stride]:
                            matrix.append(float(y))
                        self.geometry_info['joints'][x]['matrix'] = matrix

        self.geometry_info['weights'] = {}
        vertex_weights = skin.find('collada:vertex_weights', self.namespaces)
        vertex_weights_inputs = vertex_weights.findall('collada:input', self.namespaces)
        for _input in vertex_weights_inputs:
            semantic = _input.attrib['semantic']
            source_url = _input.attrib['source']
            source = skin.find(f'collada:source[@id="{source_url[1:]}"]', self.namespaces)

            if semantic == 'WEIGHT':
                accessor = source.find('collada:technique_common/collada:accessor', self.namespaces)
                accessor_source_url = accessor.attrib['source']
                accessor_source = source.find(f'collada:*[@id="{accessor_source_url[1:]}"]', self.namespaces)

                params = accessor.findall('collada:param', self.namespaces)
                for param in params:
                    param_name = param.attrib['name']
                    # param_type = param.attrib['type']

                    if param_name == 'WEIGHT':
                        weights = [float(x) for x in accessor_source.text.split()]
                        self.geometry_info['weights']['weights'] = weights

            vcount = vertex_weights.find('collada:vcount', self.namespaces).text
            vcount = [int(x) for x in vcount.split()]
            self.geometry_info['weights']['vcount'] = vcount

            v = vertex_weights.find('collada:v', self.namespaces).text
            v = [int(x) for x in v.split()]
            self.geometry_info['weights']['vertex_weights'] = v

    def parse_geometry(self, geometry):
        name = geometry.attrib['id']

        self.geometry_info['name'] = name

        mesh = geometry[0]

        triangles = mesh.findall('collada:triangles', self.namespaces)
        if triangles:
            pass
        else:
            triangles = mesh.findall('collada:polylist', self.namespaces)
        inputs = triangles[0].findall('collada:input', self.namespaces)
        for _input in inputs:
            semantic = _input.attrib['semantic']
            source_link = _input.attrib['source'][1:]
            source = mesh.find(f'*[@id="{source_link}"]')

            if semantic == 'VERTEX':
                vertices_input = source[0]
                semantic = vertices_input.attrib['semantic']
                source_link = vertices_input.attrib['source'][1:]
                source = mesh.find(f'*[@id="{source_link}"]')

            float_array = source.find('collada:float_array', self.namespaces)
            accessor = source.find('collada:technique_common/collada:accessor', self.namespaces)

            vertex_temp = [float(floating) for floating in float_array.text.split()]

            scale = max(max(vertex_temp), abs(min(vertex_temp)))
            if scale < 1:
                scale = 1
            if semantic == 'TEXCOORD':
                vertex_temp[1::2] = [1 - x for x in vertex_temp[1::2]]

            vertex = []
            for x in range(0, len(vertex_temp), len(accessor)):
                vertex.append(vertex_temp[x: x + len(accessor)])

            self.geometry_info['vertices'].append({'type': semantic,
                                                   'index': 0,
                                                   'scale': scale,
                                                   'vertex': vertex})
        for triangle in triangles:
            triangles_material = triangle.attrib['material']

            p = triangle.find('collada:p', self.namespaces)
            polygons_temp = [int(integer) for integer in p.text.split()]

            polygons = []
            for x in range(0, len(polygons_temp), len(inputs) * 3):
                temp_list = []
                for x1 in range(len(inputs)):
                    second_temp_list = []
                    for x2 in range(3):
                        second_temp_list.append(polygons_temp[x + x1 + x2])
                    temp_list.append(second_temp_list)
                polygons.append(temp_list)
            self.geometry_info['materials'].append({'name': triangles_material,
                                                    'polygons': polygons})
        self.parsed['geometries'].append(self.geometry_info)



def _(*args):
    print('[ScwUtils]', end=' ')
    for arg in args:
        print(arg, end=' ')
    print()


class Parser:
    def __init__(self, file_data: str):
        # <Variables>

        self.parsed = {'header': {'version': 2,
                                  'frame_rate': 30,
                                  'materials_file': 'sc3d/character_materials.scw'},
                       'materials': [],
                       'geometries': [],
                       'cameras': [],
                       'nodes': []}

        self.lines = file_data.split('\n')

        # <Vertices>
        self.position_temp, self.position = [], []
        self.normals_temp, self.normals = [], []
        self.texcoord_temp, self.texcoord = [], []
        # </Vertices>

        self.polygons = []

        # </Variables>

        self.parse()

    def parse(self):
        geometry_name = 'This model haven\'t a name!:( Its VERY SAD!'
        is_first_name = True
        for line in self.lines:
            items = line.split()[1:]
            if line.startswith('v '):  # POSITION
                temp_list = []
                for item in items:
                    self.position_temp.append(float(item))
                    temp_list.append(float(item))
                self.position.append(temp_list)
            elif line.startswith('vn '):  # NORMAL
                temp_list = []
                for item in items:
                    self.normals_temp.append(float(item))
                    temp_list.append(float(item))
                self.normals.append(temp_list)
            elif line.startswith('vt '):  # TEXCOORD
                if len(items) > 2:
                    items = items[:-1]
                temp_list = []
                for item in items:
                    self.texcoord_temp.append(float(item))
                    temp_list.append(float(item))
                self.texcoord.append(temp_list)
            elif line.startswith('f '):
                temp_list = []
                if len(items) > 3:
                    _('It is necessary to triangulate the model')
                    break
                for item in items:
                    second_temp_list = []
                    if len(item.split('/')) == 2:
                        _('Model have not normals or texture')
                        break
                    elif len(item.split('/')) == 1:
                        _('Model have not normals and texture')
                        break
                    for x in item.split('/'):
                        second_temp_list.append(int(x) - 1)
                    temp_list.append([second_temp_list[0], second_temp_list[2], second_temp_list[1]])
                self.polygons.append(temp_list)
            elif line.startswith('o '):
                # if not is_first_name:
                #     position_scale = self.get_vertex_scale(self.position_temp)
                #     normals_scale = self.get_vertex_scale(self.normals_temp)
                #     texcoord_scale = self.get_vertex_scale(self.texcoord_temp)
                #
                #     self.parsed['geometries'].append({
                #         'name': geometry_name,
                #         'group': '',
                #         'vertices': [
                #             {'type': 'POSITION', 'index': 0, 'scale': position_scale, 'vertex': self.position},
                #             {'type': 'NORMAL', 'index': 1, 'scale': normals_scale, 'vertex': self.normals},
                #             {'type': 'TEXCOORD', 'index': 2, 'scale': texcoord_scale, 'vertex': self.texcoord}
                #         ],
                #         'have_bind_matrix': False,
                #         'materials': [{'name': 'character_mat', 'polygons': self.polygons}]
                #     })
                #
                #     # <VariablesReset>
                #
                #     # <Vertices>
                #     self.position_temp, self.position = [], []
                #     self.normals_temp, self.normals = [], []
                #     self.texcoord_temp, self.texcoord = [], []
                #     # </Vertices>
                #
                #     self.polygons = []
                #
                #     # </VariablesReset>
                # geometry_name = line.split('o ')[0]
                if is_first_name:
                    geometry_name = line.split('o ')[0]

        position_scale = self.get_vertex_scale(self.position_temp)
        normals_scale = self.get_vertex_scale(self.normals_temp)
        texcoord_scale = self.get_vertex_scale(self.texcoord_temp)

        self.parsed['geometries'].append({
            'name': geometry_name,
            'group': '',
            'vertices': [
                {'type': 'POSITION', 'index': 0, 'scale': position_scale, 'vertex': self.position},
                {'type': 'NORMAL', 'index': 1, 'scale': normals_scale, 'vertex': self.normals},
                {'type': 'TEXCOORD', 'index': 2, 'scale': texcoord_scale, 'vertex': self.texcoord}
            ],
            'have_bind_matrix': False,
            'materials': [{'name': 'character_mat', 'polygons': self.polygons}]
        })

    @staticmethod
    def get_vertex_scale(vertex_data: list):
        vertex_scale = max(max(vertex_data), abs(min(vertex_data)))
        if vertex_scale < 1:
            vertex_scale = 1
        return vertex_scale



def matrix2tuple(matrix_list: list):
    new_matrix = [tuple(row) for row in matrix_list]
    new_matrix = tuple(new_matrix)

    return new_matrix


class Matrix2x2(object):
    def __init__(self, matrix_list: tuple = None):
        if matrix_list is not None:
            self.matrix = matrix_list
        else:
            self.matrix = (
                (1, 0),  # x
                (0, 1),  # y
            )

    def __matmul__(self, other):
        if len(self.matrix) != len(other.matrix[0]) or len(self.matrix[0]) != len(other.matrix):
            raise TypeError('Матрицы не могут быть перемножены')
        else:
            multiplied_matrix = []

            for row in self.matrix:
                matrix_row = []
                for column in range(2):
                    s = 0
                    for i in range(2):
                        s += row[i] * other.matrix[i][column]
                    matrix_row.append(s)
                multiplied_matrix.append(matrix_row)

            return Matrix2x2(matrix2tuple(multiplied_matrix))

    def __mul__(self, other: int or float):
        new_matrix = []
        for row in range(len(self.matrix)):
            new_row = []
            for column in range(len(self.matrix[0])):
                new_row.append(self.matrix[row][column]*other)
            new_matrix.append(new_row)

        return Matrix2x2(matrix2tuple(new_matrix))

    def inverse(self):
        det = self.determinant()
        if det != 0:
            self.transpose()
            self.cofactor()

            self.matrix = self.__mul__(1 / det).matrix

        return self

    def __str__(self):
        return str(self.matrix)

    def determinant(self):
        value1_1 = self.matrix[0][0]
        value1_2 = self.matrix[0][1]

        value2_1 = self.matrix[1][0]
        value2_2 = self.matrix[1][1]

        det = value1_1 * value2_2
        det -= value1_2 * value2_1

        return det

    def transpose(self):
        value1_1 = self.matrix[0][0]
        value1_2 = self.matrix[0][1]

        value2_1 = self.matrix[1][0]
        value2_2 = self.matrix[1][1]

        self.matrix = (
            (value2_2, value1_2),
            (value2_1, value1_1)
        )

    def cofactor(self):
        cofactor_matrix = []
        for row in range(len(self.matrix)):
            new_row = []
            for column in range(len(self.matrix[0])):
                new_row.append(self.matrix[row][column] * (-1) ** (row + column))
            cofactor_matrix.append(new_row)
        self.matrix = cofactor_matrix




import binascii

from ..chunks import *


def _(*args):
    print('[ScwUtils]', end=' ')
    for arg in args:
        print(arg, end=' ')
    print()


class Writer:
    def __init__(self, data: dict):
        self.writen = b'SC3D'

        head = HEAD.Encoder(data['header'])

        self.write_chunk(head)

        # TODO: make materials
        for material in data['materials']:
            mate = MATE.Encoder(material)

            self.write_chunk(mate)

        for geometry in data['geometries']:
            geom = GEOM.Encoder(geometry)

            self.write_chunk(geom)

        # TODO: make cameras
        for camera in data['cameras']:
            came = CAME.Encoder(camera)

            self.write_chunk(came)

        node = NODE.Encoder(data['nodes'])
        self.write_chunk(node)

        wend = WEND.Encoder()
        self.write_chunk(wend)

    def write_chunk(self, chunk):
        self.writen += chunk.length.to_bytes(4, 'big') + chunk.name.encode('utf-8') + chunk.buffer
        self.writen += binascii.crc32(chunk.name.encode('utf-8') + chunk.buffer).to_bytes(4, 'big')
